# Copyright 2010-2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from nose.tools import istest
from reahl.tofu import Fixture, test, vassert, expected, NoException
from reahl.stubble import EmptyStub

from reahl.web.fw import Region
from reahl.web.ui import TwoColumnPage, P
from reahl.webdev.tools import Browser
from reahl.web_dev.fixtures import WebFixture, ReahlWebApplicationStub
from reahl.component.exceptions import ProgrammerError, IncorrectArgumentError, IsSubclass, IsInstance

@istest
class AppBasicsTests(object):
    @test(WebFixture)
    def basic_assembly(self, fixture):
        """An application is built by extending ReahlWebApplication and providing an assemble method.
        
        The most basic application needs a main window and a view. Views are mapped to Urls.
        """
        class MainRegion(Region):
            def assemble(self):
                self.define_main_window(TwoColumnPage)
                self.define_view(u'/', title=u'Hello')

        webapp = fixture.new_webapp(site_root=MainRegion)
        browser = Browser(webapp)

        # GETting the URL results in the HTML for that View
        browser.open('/')
        vassert( browser.title == u'Hello' )

        # The headers are set correctly
        response = browser.last_response
        vassert( response.content_length == 893 )
        vassert( response.content_type == 'text/html' )
        vassert( response.charset == 'utf-8' )

        # Invalid URLs do not exist
        browser.open('/nonexistantview/', status=404)

    @test(WebFixture)
    def basic_error1(self, fixture):
        """Sending the the wrong kind of thing as widget_class to define_main_window is reported to the programmer."""
        class MainRegion(Region):
            def assemble(self):
                self.define_main_window(EmptyStub)
                self.define_view(u'/', title=u'Hello')

        webapp = fixture.new_webapp(site_root=MainRegion)
        browser = Browser(webapp)

        with expected(IsSubclass):
            browser.open('/')

    @test(WebFixture)
    def basic_error2(self, fixture):
        """Sending the the wrong arguments for the specified class to define_main_window is reported to the programmer."""
        class MainRegion(Region):
            def assemble(self):
                self.define_main_window(TwoColumnPage, 1, 2)
                self.define_view(u'/', title=u'Hello')

        webapp = fixture.new_webapp(site_root=MainRegion)
        browser = Browser(webapp)

        def check_exc(ex):
            msg = str(ex)
            vassert( msg.startswith('define_main_window was called with arguments that do not match those expected by') )
        with expected(IncorrectArgumentError, test=check_exc):
            browser.open('/')

    @test(WebFixture)
    def basic_error3(self, fixture):
        """Forgetting to define a main_window is reported to the programmer."""
        class MainRegion(Region):
            def assemble(self):
                self.define_view(u'/', title=u'Hello')

        webapp = fixture.new_webapp(site_root=MainRegion)
        browser = Browser(webapp)

        def check_exc(ex):
            msg = str(ex)
            vassert( msg == 'there is no main_window defined for /' )
        with expected(ProgrammerError, test=check_exc):
            browser.open('/')
            
    @test(WebFixture)
    def slots(self, fixture):
        """A View modifies the main window by populating named Slots in the main window with Widgets."""
        class MainRegion(Region):
            def assemble(self):
                self.define_main_window(TwoColumnPage)
                home = self.define_view(u'/', title=u'Hello')
                home.set_slot(u'main', P.factory(text=u'Hello world'))
                home.set_slot(u'footer', P.factory(text=u'I am the footer'))

        webapp = fixture.new_webapp(site_root=MainRegion)
        browser = Browser(webapp)
        
        browser.open('/')
        vassert( browser.title == u'Hello' )
        [main_p, footer_p] = browser.lxml_html.xpath('//p')
        vassert( main_p.text == u'Hello world' )
        vassert( footer_p.text == u'I am the footer' )

    @test(WebFixture)
    def slot_error(self, fixture):
        """Supplying contents for a slot that does not exist results in s sensible error."""
        class MainRegion(Region):
            def assemble(self):
                self.define_main_window(TwoColumnPage)
                home = self.define_view(u'/', title=u'Hello')
                home.set_slot(u'main', P.factory(text=u'Hello world'))
                home.set_slot(u'nonexistantslotname', P.factory(text=u'I am breaking'))

        webapp = fixture.new_webapp(site_root=MainRegion)
        browser = Browser(webapp)

        def check_exc(ex):
            vassert( str(ex).startswith(u'An attempt was made to plug Widgets into the following slots that do not exist') )

        with expected(ProgrammerError, test=check_exc):
            browser.open('/')


    @test(WebFixture)
    def slot_defaults(self, fixture):
        """If a View does not specify contents for a Slot, the Slot will be populated by the window's default
           widget for that slot if specified, else it will be left empty.
        """
        class MainRegion(Region):
            def assemble(self):
                main = self.define_main_window(TwoColumnPage)
                main.add_default_slot(u'main', P.factory(text=u'defaulted slot contents'))
                self.define_view(u'/', title=u'Hello')

        webapp = fixture.new_webapp(site_root=MainRegion)
        browser = Browser(webapp)
        
        browser.open('/')

        # The default widget for the main slot is used
        [p] = browser.lxml_html.xpath('//p')
        vassert( p.text == 'defaulted slot contents' )

        # The header slot has no default, and is thus left empty
        header_contents = browser.lxml_html.xpath('//header/*')
        vassert( not header_contents )



# Copyright 2011, 2012, 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from webob import Request

from nose.tools import istest
from reahl.tofu import Fixture, test
from reahl.tofu import vassert, expected
from reahl.stubble import stubclass, EmptyStub

from reahl.web.fw import Region, Widget, FactoryDict, RegionFactory, RegexPath
from reahl.web.ui import TwoColumnPage, P, A, Panel, Slot
from reahl.webdev.tools import Browser, WidgetTester
from reahl.web_dev.fixtures import WebFixture


@istest
class RegionTests(object):
    @test(WebFixture)
    def basic_region(self, fixture):
        """A Region is a chunk of web app that can be grafted onto the URL hierarchy of any app.
        
           A Region has its own views. Its Views are relative to the Region itself.
        """
        class RegionWithTwoViews(Region):
            def assemble(self):
                self.define_view(u'/', title=u'Region root view')
                self.define_view(u'/other', title=u'Region other view')

        class MainRegion(Region):
            def assemble(self):
                self.define_main_window(TwoColumnPage)
                self.define_region(u'/aregion',  RegionWithTwoViews,  {}, name=u'myregion')

        webapp = fixture.new_webapp(site_root=MainRegion)
        browser = Browser(webapp)

        browser.open('/aregion/')
        vassert( browser.title == u'Region root view' )

        browser.open('/aregion/other')
        vassert( browser.title == u'Region other view' )

    @test(WebFixture)
    def region_slots_map_to_window(self, fixture):
        """The Region uses its own names for Slots. When attaching a Region, you have to specify 
            which of the Region's Slots plug into which of the main window's Slots.
        """
        class RegionWithSlots(Region):
            def assemble(self):
                root = self.define_view(u'/', title=u'Region root view')
                root.set_slot(u'text', P.factory(text=u'in region slot named text'))

        class MainRegion(Region):
            def assemble(self):
                self.define_main_window(TwoColumnPage)
                self.define_region(u'/aregion',  RegionWithSlots,  {u'text': u'main'}, name='myregion')

        webapp = fixture.new_webapp(site_root=MainRegion)
        browser = Browser(webapp)

        browser.open('/aregion/')
        vassert( browser.title == u'Region root view' )

        # The widget in the Region's slot named 'text' end up in the TwoColumnPage slot called main
        [p] = browser.lxml_html.xpath('//div[@id="yui-main"]/div/p')
        vassert( p.text == 'in region slot named text' )


    @test(WebFixture)
    def region_redirect(self, fixture):
        """When opening an URL without trailing slash that maps to where a Region is attached,
           the browser is redirected to the Region '/' View."""
           
        class RegionWithRootView(Region):
            def assemble(self):
                self.define_view(u'/', title=u'Region root view')

        class MainRegion(Region):
            def assemble(self):
                self.define_main_window(TwoColumnPage)
                self.define_region(u'/aregion',  RegionWithRootView,  {}, name='myregion')

        webapp = fixture.new_webapp(site_root=MainRegion)
        browser = Browser(webapp)

        browser.open('/aregion')
        vassert( browser.title == u'Region root view' )
        vassert( browser.location_path == u'/aregion/' )

    @test(WebFixture)
    def region_arguments(self, fixture):
        """Regions can take exta args and kwargs."""
           
        class RegionWithArguments(Region):
            def assemble(self, kwarg=None):
                self.kwarg = kwarg
                text = self.kwarg
                root = self.define_view(u'/', title=u'A view')
                root.set_slot(u'text', P.factory(text=text))

        class MainRegion(Region):
            def assemble(self):
                self.define_main_window(TwoColumnPage)
                self.define_region(u'/aregion', RegionWithArguments, {u'text': u'main'},
                                name='myregion', kwarg=u'the kwarg')

        webapp = fixture.new_webapp(site_root=MainRegion)
        browser = Browser(webapp)

        browser.open('/aregion/')
        [p] = browser.lxml_html.xpath('//p')
        vassert( p.text == 'the kwarg' )

    @test(WebFixture)
    def bookmarks(self, fixture):
        """Bookmarks are pointers to Views. You need them, because Views are relative to a Region and
           a Bookmark can, at run time, turn these into absolute URLs. Bookmarks also contain metadata,
           such as the title of the View it points to.
        """
        class RegionWithRelativeView(Region):
            def assemble(self):
                self.define_view(u'/aview', title=u'A View title')

        class MainRegion(Region):
            def assemble(self):
                self.define_main_window(TwoColumnPage)
                region_factory = self.define_region(u'/aregion',  RegionWithRelativeView,  {}, name=u'myregion')

                # How you could get a bookmark from a RegionFactory
                fixture.bookmark = region_factory.get_bookmark(relative_path=u'/aview')

        webapp = fixture.new_webapp(site_root=MainRegion)
        Browser(webapp).open('/aregion/aview') # To execute the above once

        # What the bookmark knows
        vassert( fixture.bookmark.href.path == u'/aregion/aview' )
        vassert( fixture.bookmark.description == u'A View title' )
        vassert( fixture.bookmark.base_path == u'/aregion' )
        vassert( fixture.bookmark.relative_path == u'/aview' )

        # How you would use a bookmark in other views (possibly in other regions)
        a = A.from_bookmark(fixture.view, fixture.bookmark)
        
        # .. and how the A will be rendered
        a_etree = WidgetTester(a).render_html_tree()        
        vassert( a_etree.attrib[u'href'] == u'/aregion/aview' )
        vassert( a_etree.text == u'A View title' )

    class LifeCycleFixture(WebFixture):
        def current_view_is_plugged_in(self, main_window):
            return main_window.slot_contents[u'main_slot'].__class__ is Panel

    @test(LifeCycleFixture)
    def the_lifecycle_of_a_region(self, fixture):
        """This test illustrates the steps a Region goes through from being specified, to
           being used. It tests a couple of lower-level implementation issues (see comments)."""

        @stubclass(Region)
        class RegionStub(Region):
            assembled = False
            def assemble(self, **region_arguments):
                self.controller_at_assemble_time = self.controller
                root = self.define_view(u'/some/path', title=u'A view')
                root.set_slot(u'slotA', Panel.factory())
                self.assembled = True

        # Phase1: specifying a region and assembleing it to a site (with kwargs)
        parent_region = None
#        parent_region = EmptyStub(base_path=u'/')
        slot_map = {u'slotA': u'main_slot'}
        region_factory = RegionFactory(parent_region, RegexPath(u'/', u'/', {}), slot_map, RegionStub, u'test_region')


        # Phase2: creating a region
        request = Request.blank(u'/some/path')
        fixture.context.set_request(request)
        region = region_factory.create(u'/some/path', for_bookmark=False)

        # - Assembly happened correctly
        vassert( region.parent_region is parent_region )
        vassert( region.slot_map is slot_map )
        vassert( region.name is u'test_region' )
        vassert( region.relative_base_path == u'/' )
        vassert( region.controller_at_assemble_time is not None)
        vassert( region.controller is not None )
        vassert( region.assembled )

        # - Create for_bookmark empties the relative_path
        region = region_factory.create(u'/some/path', for_bookmark=True)
        vassert( region.relative_path == u'' )

        # - The relative_path is reset if not done for_bookmark
        #   This is done in case a for_bookmark call just
        #   happened to be done for the same Region in the same request
        #   before the "real" caal is done
        region = region_factory.create(u'/some/path', for_bookmark=False)
        vassert( region.relative_path == u'/some/path' )

        # Phase5: create the main_window and plug the view into it
        main_window = Widget.factory().create(region.current_view)
        main_window.add_child(Slot(region.current_view, 'main_slot'))

        main_window.plug_in(region.current_view)
        vassert( fixture.current_view_is_plugged_in(main_window) )
        vassert( isinstance(region.sub_resources, FactoryDict) )



# Copyright 2011, 2012, 2013 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.




from nose.tools import istest
from reahl.tofu import Fixture, test, scenario
from reahl.tofu import vassert, expected

from reahl.component.modelinterface import Field, IntegerField, exposed, Event
from reahl.web.fw import ReahlWebApplication, Region, UrlBoundView, IdentityDictionary, CannotCreate
from reahl.web.ui import TwoColumnPage, P, A, Form, Button
from reahl.webdev.tools import Browser, WidgetTester, XPath
from reahl.web_dev.fixtures import WebFixture, ReahlWebApplicationStub


@istest
class ParameterisedTests(object):
    class ParameterisedScenarios(WebFixture):
        @scenario
        def normal_arguments(self):
            """Arguments can be sent from where the View is defined."""
            self.argument = u'some arg'
            self.expected_value = u'some arg'
            self.url = u'/aregion/aview'
            self.should_exist = True
        
        @scenario
        def url_arguments(self):
            """Arguments can be parsed from an URL, iff they are specified to the definition as Fields."""
            self.argument = Field()
            self.expected_value = u'test1'
            self.url = u'/aregion/aview/test1'
            self.should_exist = True
            
        @scenario
        def cannot_create(self):
            """To indicate that a view does not exist for the given arguments, the .assemble() 
               method of the View should raise CannotCreate()."""
            self.argument = u'doesnotexist'
            self.url = u'/aregion/aview'
            self.should_exist = False

    @test(ParameterisedScenarios)
    def views_with_parameters(self, fixture):
        """Views can have arguments that originate from code, or are parsed from the URL."""

        class ParameterisedView(UrlBoundView):
            def assemble(self, some_arg=None):
                if some_arg == u'doesnotexist':
                    raise CannotCreate()
                self.title = u'View for: %s' % some_arg
                self.set_slot(u'main', P.factory(text=u'content for %s' % some_arg))

        class RegionWithParameterisedViews(Region):
            def assemble(self):
                self.define_view(u'/aview', view_class=ParameterisedView, some_arg=fixture.argument) 

        class MainRegion(Region):
            def assemble(self):
                self.define_main_window(TwoColumnPage)
                self.define_region(u'/aregion',  RegionWithParameterisedViews,  {u'main': u'main'}, name=u'myregion')

        webapp = fixture.new_webapp(site_root=MainRegion)
        browser = Browser(webapp)

        if fixture.should_exist:
            browser.open(fixture.url)
            vassert( browser.title == u'View for: %s' % fixture.expected_value )
            vassert( browser.is_element_present(XPath.paragraph_containing(u'content for %s' % fixture.expected_value)) )
        else:
            browser.open(fixture.url, status=404)


    @test(WebFixture)
    def views_from_regex(self, fixture):
        """Parameterised Views can also be added based on a regex over the url."""

        class ParameterisedView(UrlBoundView):
            def assemble(self, some_key=None):
                self.title = u'View for: %s' % some_key
            
        class RegionWithParameterisedViews(Region):
            def assemble(self):
                self.define_regex_view(u'/someurl_(?P<some_key>.*)', u'/someurl_${some_key}', view_class=ParameterisedView,
                                       some_key=Field(required=True))

        class MainRegion(Region):
            def assemble(self):
                self.define_main_window(TwoColumnPage)
                self.define_region(u'/aregion',  RegionWithParameterisedViews,  {}, name=u'myregion')

        webapp = fixture.new_webapp(site_root=MainRegion)
        browser = Browser(webapp)

        # Parameterisedally constructing a View from an URL
        browser.open('/aregion/someurl_test1')
        vassert( browser.title == u'View for: test1' )

    @test(WebFixture)
    def regions_from_regex(self, fixture):
        """Sub Regions can be created on the fly on a Region, based on the URL visited. To indicate that a
           Region does not exist, the creation method should return None."""

        class RegexRegion(Region):
            def assemble(self, region_key=None):
                if region_key == u'doesnotexist':
                    raise CannotCreate()

                self.name = u'region-%s' % region_key
                root = self.define_view(u'/', title=u'Simple region %s' % self.name)
                root.set_slot(u'region-slot', P.factory(text=u'in region slot'))

        class RegionWithParameterisedRegions(Region):
            def assemble(self):
                self.define_regex_region(u'/apath/(?P<region_key>[^/]*)',
                                         u'/apath/${region_key}',
                                         RegexRegion,
                                         {u'region-slot': u'main'},
                                         region_key=Field())

        class MainRegion(Region):
            def assemble(self):
                self.define_main_window(TwoColumnPage)
                self.define_region(u'/aregion',  RegionWithParameterisedRegions,  IdentityDictionary(), name=u'myregion')

        webapp = fixture.new_webapp(site_root=MainRegion)
        browser = Browser(webapp)

        # A sub-region is dynamically created from an URL
        browser.open('/aregion/apath/test1/')
        vassert( browser.title == u'Simple region region-test1' )

        # The slots of the sub-region is correctly plugged into the main window
        [p] = browser.lxml_html.xpath('//p')
        vassert( p.text == u'in region slot' )

        # Another sub-region is dynamically created from an URL
        browser.open('/aregion/apath/another/')
        vassert( browser.title == u'Simple region region-another' )

        # When the URL cannot be mapped
        browser.open('/aregion/apath/doesnotexist/', status=404)


    class ParameterisedRegionScenarios(WebFixture):
        @scenario
        def normal_arguments(self):
            """Arguments can be sent from where the Region is defined."""
            self.argument = u'some arg'
            self.expected_value = u'some arg'
            self.url = u'/aregion/parameterisedregion/aview'
            self.should_exist = True
        
        @scenario
        def url_arguments(self):
            """Arguments can be parsed from an URL, iff they are specified to the definition as Fields."""
            self.argument = Field()
            self.expected_value = u'test1'
            self.url = u'/aregion/parameterisedregion/test1/aview'
            self.should_exist = True
            
        @scenario
        def cannot_create(self):
            """To indicate that a Region does not exist for the given arguments, the .assemble() 
               method of the Region should raise CannotCreate()."""
            self.argument = u'doesnotexist'
            self.url = u'/aregion/parameterisedregion/aview'
            self.should_exist = False

    @test(ParameterisedRegionScenarios)
    def parameterised_regions(self, fixture):
        """Sub Regions can also be parameterised by defining arguments in .define_region, and receiving them in .assemble()."""

        class ParameterisedRegion(Region):
            def assemble(self, region_arg=None):
                if region_arg == u'doesnotexist':
                    raise CannotCreate()

                self.name = u'region-%s' % region_arg
                root = self.define_view(u'/aview', title=u'Simple region %s' % self.name)
                root.set_slot(u'region-slot', P.factory(text=u'in region slot'))


        class RegionWithParameterisedRegions(Region):
            def assemble(self):
                self.define_region(u'/parameterisedregion', ParameterisedRegion, {u'region-slot': u'main'}, 
                                   region_arg=fixture.argument,
                                   name=u'paramregion')

        class MainRegion(Region):
            def assemble(self):
                self.define_main_window(TwoColumnPage)
                self.define_region(u'/aregion',  RegionWithParameterisedRegions,  IdentityDictionary(), name=u'myregion')

        webapp = fixture.new_webapp(site_root=MainRegion)
        browser = Browser(webapp)

        if fixture.should_exist:
            browser.open(fixture.url)
            
            # The correct argument was passed
            vassert( browser.title == u'Simple region region-%s' % fixture.expected_value )

            # The slots of the sub-region is correctly plugged into the main window
            [p] = browser.lxml_html.xpath('//p')
            vassert( p.text == u'in region slot' )
        else:
            # When the URL cannot be mapped
            browser.open(fixture.url, status=404)




